# This script sets up a simulation environment to test streaming media 
# over 802.11 wireless interfaces with adjustable link quality.
# A dumbbell topology with wireless nodes on one end and media servers
# on the opposite end will be demonstrated.
########################################################################
# No rights reserved: Torgeir Haukaas 2007

# ======================================================================
# General Parameters
# ======================================================================
set opt(title) 		"802.11"	;# script-name
set opt(seed)		13.37		;# general random seed for simulation traffic flow
set opt(stop)		55.0		;# simulation time
set opt(verbose)	1		;# 1 to print messages 0 to be silent
# ======================================================================
# Channel properties (wifi-link)
# ======================================================================
set opt(chan)		Channel/WirelessChannel
set opt(prop)		Propagation/TwoRayGround ;# FreeSpace, TwoRayGround or Shadowing
set opt(netif)		Phy/WirelessPhy
set opt(mac)		Mac/802_11
set opt(qm)		DropTail	;# RED
set opt(ifq)		Queue/DropTail
set opt(ll)		LL
set opt(ant)            Antenna/OmniAntenna
set opt(adhocRouting)	DSDV		;# possible ad-hoc routing protocols:
					# DSDV DSR TORA AODV
set opt(disableRTS_CTS) 1		;# RTS / CTS is disabled by default (=1). ENABLE (=0) if mobile nodes
					# cannot see eachother!!
# For Propagation/Shadowing:
set opt(pathlossExp)	2.0		;# path loss exponent
set opt(std_db)		4.0		;# deviation (dB)
set opt(dist0)		1.0		;# reference distance (m) (see ns manual p. 189)
# ======================================================================
# Topology
# ======================================================================
set opt(x)		670		;# X dimension of the topography
set opt(y)		670		;# Y dimension of the topography

set opt(ifqlen)		50		;# max packet in ifq
set opt(nofmn)		1		;# number of wireless nodes OBS! max 9 for now
set opt(nofbs)		1		;# number of base stations
set opt(nofsn)		1		;# number of server nodes
set opt(mn_motion)	0		;# random motion of mobile nodes
set opt(nodeDist)	2		;# distance in meters between WLAN nodes
set opt(ecn) 		0		;# 1 to enable ECN in routers
set opt(adaptive)	true
set opt(minth)		0
set opt(maxth)		0
# ======================================================================
# Evalvid parameters
# ======================================================================
set opt(videoname)	"3xCoastguard_qcif"	;# base of video file name
set opt(vbr_rate)	3Mb			;# max VBR-output
set opt(q_variants)	30			;# quantizer scale range
set opt(ra_mtu)		1000			;# max fragmented size, MTU
set opt(fps)		30			;# frames per second (all video is treated with equal fps)
set opt(tfrc_pktsize)	[expr $opt(ra_mtu) + 36] ;# DCCP/TFRC header (12+4=16B) and IP-header (20B)
set opt(udp_pktsize)	[expr $opt(ra_mtu) + 28] ;# UDP header (8B) and IP-header (20B)
set opt(cbr_true)	0
set opt(GoP_size)	12
# ======================================================================
# Traffic generation.
# ======================================================================
set opt(nofTCP)		0	;# number of TCP flows
set opt(nofWEB)		0	;# number of WEB flows
set opt(nofFTP)		0	;# number of FTP flows
set opt(nofUDP)		0	;# number of UDP flows
set opt(nofTFRC)	0	;# number of TFRC flows
#
set opt(window)		30	;# window for long-lived traffic
set opt(tcp_pktsize)	1460
set opt(tcpTick_)	0.01
set opt(doPing)		0	;# 1 to ping servers on startup, 0 to disable
# WEB-settings
set opt(webSeed)	0	;# seed for RNG for web traffic generation
set opt(interPage)	1	;# average value for exponentially distributed inter page time
set opt(pageSize)	3	;# nof objects per page
set opt(nofWebPages)	10	;# number of pages per session
set opt(nofWebSessions)	4	;# number of web sessions
set opt(webObjectSize)	20	;# average size of web object in pkts.
set opt(paretoShape)	1.05	;# shape parameter for Pareto distribution of web size
			 	 # a larger parameter means more small objects
# ======================================================================
# Plotting statistics.
# ======================================================================
set opt(out_dir)	nsout
set opt(tf)		$opt(out_dir)/$opt(title).tr	;# trace file
set opt(namtf)		$opt(out_dir)/$opt(title).nam	;# nam trace file
set opt(doTrace)	1	;# 1 to do tracing
set opt(nam)		1	;# 1 to trace and run nam
set opt(sampleTime)	0.5	;# time between each bw-sample written by the record-function
set opt(macTrace)	OFF
set opt(agentTrace)	ON
set opt(routerTrace)	OFF
set opt(movementTrace)	OFF
# ======================================================================
# Link characteristics.
# ======================================================================
set opt(errUnit)	packet	; # unit of errors, per packet or per bit
set opt(errRateUL)	0	; # the rate of errors in uplink 
set opt(errBurstUL)	0	; # coefficient 0..1 of how bursty errors are
set opt(errSlotUL)	0	; # time in sec of low error period
set opt(errRateDL)	0	; # 
set opt(errBurstDL)	0	; # 
set opt(errSlotDL)	0
# ======================================================================
set opt(tfrcFB)		1		; # number of feedback reports in TFRC per RTT

set opt(snet_bw)	100Mb		;# link BW on server access net
set opt(snet_prop)	2ms		;# link propagation delay on server access net
set opt(core_bw)	100Mb		;# link BW on core net (router0<--->router1)
set opt(core_prop)	2ms		;# link propagation delay on core net (router0<--->router1)
set opt(bsnet_bw)	100Mb		;# link BW on base station access net
set opt(bsnet_prop)	2ms		;# link propagation delay on base station access net
set opt(wifi_bw)	11Mb		;# link BW on wifi net

# ==================== END OPTIONS =====================================
# ==================== START PROCEDURES ================================
proc getopt {argc argv} {
        global opt
        for {set i 0} {$i < $argc} {incr i} {
		set arg [lindex $argv $i]
		if {[string range $arg 0 0] != "-"} continue
		set name [string range $arg 1 end]
		set opt($name) [lindex $argv [expr $i+1]]
		puts2 "opt($name): $opt($name)" 
        }
}

proc puts2 {output} {
	global opt
	if { $opt(verbose)==1 } {
		puts $output
	}
}
proc setnofnodes {} {
	global opt ns
	set opt(nofmn) [expr $opt(nofTFRC)+$opt(nofUDP)+$opt(nofFTP)+$opt(nofWEB)]

	if { $opt(nofmn)>9 } {
		puts2 "nofmn>9. extend adress array!"
		puts2 "NS EXITING..."
		$ns halt
		exit 0
	}
	set opt(nofsn) 0
	if {$opt(nofTFRC)>0 } {
		set opt(nofsn) [expr $opt(nofsn) + $opt(nofTFRC)]
	}
	if {$opt(nofUDP)>0 } {
		incr opt(nofsn)
	}
	if {$opt(nofFTP)>0 } {
		incr opt(nofsn)
	}
	if {$opt(nofWEB)>0 } {
		incr opt(nofsn)
	}
	puts2 "Ordered one server per source type. nof servers: $opt(nofsn)"
}

proc wlan_topo_complex {} {
	global ns nodes opt mn sn rn bs
	puts2 "Building topology..."
	########################################################
	#  Server,       Router,  Base-station and Mobile nodes
	#
	#  sn_0---\                             /----mn_0
	#          \                           /
	#  sn_1-----rn_0---------rn_1------bs_0------mn_1
	#           /              \          \       ...
	#  ...     /                \   ....   \
	#         /                  \          \----mn_n
	#  sn_n--/                    \----bs_n
	#
	#  0.0.n    0.0.0        1.0.0    1.n.0      1.n.m
	########################################################

	# Configuration for Orinoco 802.11b 11Mbps PC card with ->22.5m range
	Phy/WirelessPhy set Pt_ 0.031622777		;# transmit power
	Phy/WirelessPhy set L_ 1.0			;# System loss factor
	Phy/WirelessPhy set bandwidth_ 1 ;#$opt(wifi_bw)
	Phy/WirelessPhy set freq_ 2.472e9 		;# channel-13. 2.472GHz
	Phy/WirelessPhy set CPThresh_ 10.0		;# reception of simultaneos packets if P1 > CPThresh_ x P2
	Phy/WirelessPhy set CSThresh_ 5.011872e-12	;# carrier sensing threshold
	Phy/WirelessPhy set RXThresh_ 5.82587e-09	;# reception threshold

	# LIMITATION: ns does not support ARF (Auto Rate Fallback, dynamic channel rate adjustment)
	Mac/802_11 set dataRate_ $opt(wifi_bw)

	# Most commercial 802.11b-cards have RTS/CTS disabled by default. RTSThreshold_ = 3000 will
	# enable RTS only for packets above 3KB, which should be never.
	if { $opt(disableRTS_CTS)==1 } {
		puts2 "Disabling RTS/CTS"
		Mac/802_11 set RTSThreshold_ 3000
	}
	Mac/802_11 set basicRate_ 1Mb			;# for broadcast packets

	# address scheme set-up
	$ns node-config -addressType hierarchical
	AddrParams set domain_num_ 2		;# sets two domains: 0.0.0 and 1.0.0
	if { $opt(nofbs)>1 } {
		puts2 "no support for more than one base station node"
		puts2 "NS EXITING..."
		$ns halt
		exit 0
	}
	lappend cluster_num  1 2		;# sets nof clusters pr domain (2 = one for router, one for each bs)
						;# one cluster of servers and one cluster of mobile nodes
	AddrParams set cluster_num_ $cluster_num
	# setting nof nodes pr cluster [clients+router], router, [clients+base station]:
	lappend eilastlevel [expr $opt(nofsn) + 1] 1 [expr $opt(nofmn) + 1]
	AddrParams set nodes_num_ $eilastlevel
	
	set topo [new Topography]
	$topo load_flatgrid $opt(x) $opt(y)
	# god needs to know the number of all wireless interfaces MN+BS
	create-god [expr $opt(nofmn) + $opt(nofbs)]

	# creating wired nodes: sn = server nodes and rn = router nodes
	set rn(0) [$ns node {0.0.0}]
	set rn(1) [$ns node {1.0.0}]
	puts2 "Router nodes:"
	puts2 [$rn(0) node-addr]
	puts2 [$rn(1) node-addr]
	# create more sn-adresses if necessary...
	if { $opt(nofsn)>9 } {
		puts2 "create for-loop to address more server nodes"
		puts2 "NS EXITING..."
		$ns halt
	}
	set sn_adr {0.0.1 0.0.2 0.0.3 0.0.4 0.0.5 0.0.6 0.0.7 0.0.8 0.0.9}
	puts2 "Server nodes:"
	for {set i 0 } { $i < $opt(nofsn) } { incr i } {
		set sn($i) [$ns node [lindex $sn_adr $i]]
		puts2 [$sn($i) node-addr]
	}

	if {$opt(prop) == "Propagation/Shadowing" } {
		# first set values of shadowing model
		Propagation/Shadowing set pathlossExp_	$opt(pathlossExp)	;# path loss exponent
		Propagation/Shadowing set std_db_	$opt(std_db)		;# shadowing deviation (dB)
		Propagation/Shadowing set dist0_ 	$opt(dist0)		;# reference distance (m)
		Propagation/Shadowing set seed_ 	$opt(seed)		;# seed for RNG
		puts2 "values for shadowing propagation are set"
	}
	# configuring node for 802.11-interface.
	$ns node-config -adhocRouting $opt(adhocRouting) \
		-llType $opt(ll) \
		-macType $opt(mac) \
		-ifqType $opt(ifq) \
		-ifqLen $opt(ifqlen) \
		-propInstance [new $opt(prop) ] \
		-antType $opt(ant) \
		-phyType $opt(netif) \
		-wiredRouting ON \
		-channel [new $opt(chan)] \
		-topoInstance $topo \
		-agentTrace $opt(agentTrace) \
		-routerTrace $opt(routerTrace) \
		-macTrace $opt(macTrace) \
		-movementTrace $opt(movementTrace)
 # -propType $opt(prop) \

	# create base-station
	if { $opt(nofbs)>1 } {
		puts2 "create for-loop to add several base-stations"
		puts2 "NS EXITING..."
		$ns halt
		exit 0
	} else {
		# creating and setting network hier-address
		set bs(0) [$ns node {1.1.0}]
		# bs is not moving:
		$bs(0) random-motion 0
		$bs(0) set X_ 1.0
		$bs(0) set Y_ 1.0
		$bs(0) set Z_ 0.0
		puts2 "Base station nodes:"
		puts2 [$bs(0) node-addr]
	}
	
	# configure for "nofmn" mobilenodes
	$ns node-config -wiredRouting OFF
	if { $opt(nofbs)>1 } {
		puts2 "create for-loop to address mobile nodes"
		puts2 "under several base-stations"
		puts2 "NS EXITING..."
		$ns halt
		exit 0
	} else {
		set x_pos 0.0
		set y_pos 0.0
		# create more mn-adresses if necessary...
		if { $opt(nofmn)>9 } {
			puts2 "create for-loop to address more mobile nodes"
			puts2 "NS EXITING..."
			$ns halt
			exit 0
		}
		set adr {1.1.1 1.1.2 1.1.3 1.1.4 1.1.5 1.1.6 1.1.7 1.1.8 1.1.9}
		puts2 "Mobile nodes: "
		for {set j 0 } { $j < $opt(nofmn) } { incr j } {
			set mn($j) [$ns node [lindex $adr $j]]
			puts2 [lindex $adr $j]
			# setting coordinates and motion
			$mn($j) random-motion $opt(mn_motion)
			$mn($j) set X_ $opt(nodeDist) ;#[expr $x_pos + $opt(nodeDist)/2]
			$mn($j) set Y_ $opt(nodeDist) ;#[expr $y_pos + $opt(nodeDist)/2]
			$mn($j) set Z_ 0.0
			# each new mn is longer away from bs:
			set x_pos [expr $x_pos + $opt(nodeDist)/2]
			set y_pos [expr $y_pos + $opt(nodeDist)/2]
			$mn($j) base-station [AddrParams addr2id [$bs(0) node-addr]]
			$ns initial_node_pos $mn($j) 20
		}
	}

	# linking up servers to first core router
	set link_angle 135
	set angle_offset [expr 90/$opt(nofsn)]
	for {set k 0 } { $k < $opt(nofsn) } { incr k } {
		$ns duplex-link $sn($k) $rn(0) $opt(snet_bw) $opt(snet_prop) $opt(qm)
		#$ns duplex-link-op $rn(0) $sn($k) orient $link_angle
		set link_angle [expr $link_angle + $angle_offset]
	}
	# core link
	$ns duplex-link $rn(0) $rn(1) $opt(core_bw) $opt(core_prop) $opt(qm)
	#$ns duplex-link-op $rn(0) $rn(1) orient right

	# linking up base stations to second core router
	for {set l 0 } { $l < $opt(nofbs) } { incr l } {
		$ns duplex-link $rn(1) $bs($l) $opt(bsnet_bw) $opt(bsnet_prop) $opt(qm)
		#$ns duplex-link-op $rn(1) $bs($l) orient right-up
	}

	puts2 "done setting up topology"
}

proc ping {} {
	global opt ns mn sn
	set tmp 0
	set tmp2 0
	# establish route with ping. each type to a seperate server:
	for {set m $tmp2 } { $m < $opt(nofFTP) } { incr m} {
		set ping_downstream [$ns create-connection Ping $sn($tmp) Ping $mn($m) 9]
		set ping_upstream [$ns create-connection Ping $mn($m) Ping $sn($tmp) 9]
		puts2 "FTP-link: ping from [$sn($tmp) node-addr] to [$mn($m) node-addr] and back!"
		$ping_downstream oneway
		$ping_upstream oneway
		$ns after 0.1 "$ping_downstream send"
		$ns after 0.2 "$ping_upstream send"
		if {$m == [expr $opt(nofFTP) - 1] } {
			incr tmp
		}
		incr tmp2
	}

	set tmp3 $tmp2
	for {set m $tmp2 } { $m < $opt(nofWEB) } { incr m} {
		set ping_downstream [$ns create-connection Ping $sn($tmp) Ping $mn($m) 9]
		set ping_upstream [$ns create-connection Ping $mn($m) Ping $sn($tmp) 9]
		puts2 "WEB-link: ping from [$sn($tmp) node-addr] to [$mn($m) node-addr] and back!"
		$ping_downstream oneway
		$ping_upstream oneway
		$ns after 0.3 "$ping_downstream send"
		$ns after 0.4 "$ping_upstream send"
		if {$m == [expr $opt(nofWEB) - 1] } {
			incr tmp
		}
		incr tmp2
	}

	set tmp3 $tmp2
	for {set m $tmp2 } { $m < [expr $opt(nofUDP) + $tmp3] } { incr m} {
		set ping_downstream [$ns create-connection Ping $sn($tmp) Ping $mn($m) 9]
		set ping_upstream [$ns create-connection Ping $mn($m) Ping $sn($tmp) 9]
		puts2 "UDP-link: ping from [$sn($tmp) node-addr] to [$mn($m) node-addr] and back!"
		$ping_downstream oneway
		$ping_upstream oneway
		$ns after 0.5 "$ping_downstream send"
		$ns after 0.6 "$ping_upstream send"
		if {$m == [expr $opt(nofUDP) + $tmp3 - 1] } {
			incr tmp
		}
		incr tmp2
	}
	set tmp3 $tmp2
	for {set m $tmp2 } { $m < [expr $opt(nofTFRC) + $tmp3] } { incr m} {
		set ping_downstream [$ns create-connection Ping $sn($tmp) Ping $mn($m) 9]
		set ping_upstream [$ns create-connection Ping $mn($m) Ping $sn($tmp) 9]
		puts2 "TFRC-link: ping from [$sn($tmp) node-addr] to [$mn($m) node-addr] and back!"
		$ping_downstream oneway
		$ping_upstream oneway
		$ns after 0.7 "$ping_downstream send"
		$ns after 0.8 "$ping_upstream send"
		if {$m == [expr $opt(nofTFRC) + $tmp3 - 1] } {
			incr tmp
		}
		incr tmp2
	}
}

proc stop {} {
	global ns opt mn sn rn bs

	for {set j 0 } { $j < $opt(nofmn) } { incr j } {
		$ns at $opt(stop).0000010 "$mn($j) reset";
	}
	for {set i 0 } { $i < $opt(nofsn) } { incr i } {
		$ns at $opt(stop).0000010 "$sn($i) reset";
	}
	$ns at $opt(stop).1 "puts2 \"NS EXITING...\" ; $ns halt"

	exit 0
}

proc record { sink bwt } {
	global opt
	#Get an instance of the simulator
	set ns [Simulator instance]
	#Set the time after which the procedure should be called again
	set time $opt(sampleTime)
	#How many bytes have been received by the traffic sink?
	set bw [$sink set bytes_]
	#Get the current time
	set now [$ns now]
	#Open a file to write bandwidth-trace
	#Calculate the bandwidth (in MBit/s) and write it to the file
	puts $bwt "$now [expr $bw/$time*8/1000000]"
	#Reset the bytes_ values on the traffic sinks
	$sink set bytes_ 0
	#Re-schedule the procedure
	$ns at [expr $now+$time] "record $sink $bwt"
}

proc prepare_evalvid_trace {} {
	global ns opt trace_file trace_file_id
	puts2 "------------------------trace prepare start------------------------------"
	puts2 "Start making GOP and frame trace files for $opt(q_variants) rate variants"
	for {set i 1} {$i <= $opt(q_variants)} {incr i} {
		set original_file_name($i) nsin/st_$opt(videoname).yuv_Q[expr $i + 1].txt 
		set original_file_id($i) [open $original_file_name($i) r]
	}
	set trace_file_name $opt(out_dir)/video2.dat
	set trace_file_id [open $trace_file_name w]
	set trace_file [new vbrTracefile2]
	$trace_file filename $trace_file_name
	set frame_count 0
	set last_time 0
	
	# AL: toggle between multiple input files!
	# set original_file_id $original_file_id(1)
	set source_select 1
		
	set frame_size_file	$opt(out_dir)/frame_size.dat
	set gop_size_file	$opt(out_dir)/gop_size.dat
	set frame_size_file_id	[open $frame_size_file w]
	set gop_size_file_id	[open $gop_size_file w]

	for {set i 1} {$i <= $opt(q_variants)} {incr i} {
		set frame_size($i) 0
		set gop_size($i) 0
	}
	set gop_numb 0
	# Convert ASCII sender file on frame size granularity to ns-2 adapted internal format
	while {[eof $original_file_id(1)] == 0} {
		for {set i 1} {$i <= $opt(q_variants)} {incr i} {
			gets $original_file_id($i) current_line($i)
		}
		
		scan $current_line(1) "%d%s%d%s%s%s%d%s" no_ frametype_ length_ tmp1_ tmp2_ tmp3_ tmp4_ tmp5_
		#puts2 "$no_ $frametype_ $length_ $tmp1_ $tmp2_ $tmp3_ $tmp4_ $tmp5_"
		
		set time [expr 1000 * 1000/$opt(fps)] ;# Note that this is time between frames in number of us
		
		if { $frametype_ == "I" } {
			set type_v 1
			set time 0
		}
		if { $frametype_ == "P" } {
			set type_v 2
		}
		if { $frametype_ == "B" } {
			set type_v 3
		}
		
		# Write to GOP size file after each H-frame found:
		if { $frametype_ == "H" } {
			set puts_string "$gop_numb"
			for {set i 1} {$i <= $opt(q_variants)} {incr i} {
				set puts_string "$puts_string $gop_size($i)" 
			}
			puts $gop_size_file_id $puts_string
			incr gop_numb
			set type_v 0 ;# Must have different type than I-frame so that the LB(r,b) algorithm finds it!
		}
		# Write to frame_size.dat:
		set puts_string "$no_"
		for {set i 1} {$i <= $opt(q_variants)} {incr i} {
			set puts_string "$puts_string $gop_size($i)" 
		}
		puts $frame_size_file_id $puts_string
		
		for {set i 1} {$i <= $opt(q_variants)} {incr i} {
			scan $current_line($i) "%d%s%d%s%s%s%d%s" no_ frametype_ length($i) tmp1_ tmp2_ tmp3_ tmp4_ tmp5_
			set gop_size($i) [expr $gop_size($i) + $length($i) ]
		}
		
		# Write to video2.dat:
		set puts_string "$time $length_ $type_v $opt(ra_mtu)"
		for {set i 2} {$i <= $opt(q_variants)} {incr i} {
			set puts_string "$puts_string $length($i)"
		}
		puts  $trace_file_id $puts_string
		if { $frametype_ != "H" } {
			incr frame_count
		}
	}
	puts2 "#of frames written to GOP and Frame trace files: $frame_count"
	#close $original_file_id
	close $trace_file_id  ;# Note that this new trace file is closed for writing and 
	# opened below for reading through being a new Tracefile in eraTraceFile2::setup()
	puts2 "------------------------trace prepare stop-------------------------------"
}

proc attach_FTP_traffic { startnode endnode connection_number} {
	global ns opt tcp tcp_in_use ftp_server
	$startnode color red
	$endnode color red
	set start_id [$startnode id]
	set end_id [$endnode id]
	set tcp($connection_number) [new Agent/TCP]
	$tcp($connection_number) set fid_ 40
	set tcpsink($connection_number) [new Agent/TCPSink]
	$ns attach-agent $startnode $tcp($connection_number)
	$ns attach-agent $endnode $tcpsink($connection_number)
	$ns connect $tcp($connection_number) $tcpsink($connection_number)
	set ftp_server($connection_number) [new Application/FTP]
	$ftp_server($connection_number) attach-agent $tcp($connection_number)

	set bwt [open $opt(out_dir)/bwt.FTP.$connection_number.tr w] ;#$opt(bwtDelimiter).
	$ns at 0.0 "record $tcpsink($connection_number) $bwt"
	puts2 "added TCP connection number $connection_number  between nodes $start_id and $end_id"
}
# ======================================================================
# Web traffic generator by Sally Floyd (from web.tcl in ns tcl/ex) 
# Modified by TH
# ======================================================================
proc attach_WEB_traffic {startnode endnode connection_number} {
	global ns opt pool

	$pool set-server 0 $startnode
	$pool set-client $connection_number $endnode

	set start_id [$startnode id]
	set end_id [$endnode id]
	$pool set-num-session $opt(nofWebSessions)
	for {set i 0} {$i < $opt(nofWebSessions)} {incr i} {
		set interPage [new RandomVariable/Exponential]
		$interPage set avg_ $opt(interPage)
		set pageSize [new RandomVariable/Constant]
		$pageSize set val_ $opt(pageSize)
		set interObj [new RandomVariable/Exponential]
		$interObj set avg_ [expr 0.01]
		set objSize [new RandomVariable/ParetoII]
		$objSize set avg_ $opt(webObjectSize)
		$objSize set shape_ $opt(paretoShape)
		set tmp [new RandomVariable/Uniform]
		set startTime [expr 1.0 * [$tmp value] * $opt(stop) ]
		# Poisson start times for sessions.
		# when does a web session end?
		puts [format "WEB session $i startTime %0.3f" $startTime ]
		$pool create-session $i $opt(nofWebPages) $startTime $interPage $pageSize $interObj $objSize
		# All of the web traffic uses the same packet size.
		# webcache/webtraf.cc
		# doc/webcache.tex
		# Feldmann99a, Section 2.4, paragraph 3-4 and the appendix A.1.
		# empweb/empftp.cc
		# doc/session.tex
		# When does each web session start?
	}
	# using web-throughput.awk to log throughput
	puts2 "added WEB connection number $connection_number with $opt(nofWebSessions) sessions  between nodes $start_id and $end_id"
}

proc attach_UDP_traffic { startnode endnode connection_number} {
	global ns opt udp vbr_in_use vbr trace_file
	$startnode color blue
	$endnode color blue
	set start_id [$startnode id]
	set end_id [$endnode id]
	set udp($connection_number) [new Agent/eraUDP]
	$udp($connection_number) set fid_ $connection_number
	$ns attach-agent $startnode $udp($connection_number)
	set vbr($vbr_in_use) [new Application/Traffic/eraVbrTrace]
	$vbr($vbr_in_use) attach-agent $udp($connection_number)
	
	$udp($connection_number) set packetSize_ $opt(udp_pktsize)
	$udp($connection_number) set TOS_field_ 1     ;# New 120905: tag ECF enabled sources! 0 is default.
	# Limit the number of send trace files:
	if {$connection_number < 15} {
		# Connect a file name to UDP source to write transmit trace data
		$udp($connection_number) set_filename $opt(out_dir)/sd_be.udp.$connection_number 
	}
	$vbr($vbr_in_use) set running_ 0
	$vbr($vbr_in_use) set r_ $opt(vbr_rate) ;# Set the rate instead of packet interval
	$vbr($vbr_in_use) attach-tracefile $trace_file		;# video2.dat type vbrTracefile2
	$vbr($vbr_in_use) set b_ 1.5
	$vbr($vbr_in_use) set q_ 1
	$vbr($vbr_in_use) set GoP_ $opt(GoP_size)
	$vbr($vbr_in_use) set fps_ $opt(fps)
	$vbr($vbr_in_use) set isTFRC_ 0

	set udpsink($connection_number) [new Agent/eraUdpSink]
	$ns attach-agent $endnode  $udpsink($connection_number)
	$ns connect $udp($connection_number) $udpsink($connection_number)
	if {$connection_number == 0} {
		# Connect a file name to TFRC sink to write receivce trace data
		$udpsink($connection_number) set_trace_filename $opt(out_dir)/rd_be.udp.$connection_number
	}
	$udpsink($connection_number) set bytes_ 0
	set bwt [open $opt(out_dir)/bwt.UDP.$connection_number.tr w] ;#$opt(bwtDelimiter).
	$ns at 0.0 "record $udpsink($connection_number) $bwt"
	puts2 "added UDP connection number $connection_number (vbr($vbr_in_use)) between nodes $start_id and $end_id"
	incr vbr_in_use
}

# ======================================================================
# Add Evalvid-RA RA-SVBR sources
# ======================================================================
proc attach_TFRC_traffic {startnode endnode connection_number} {
	global ns tfrc opt trace_file vbr_in_use vbr
	$startnode color green
	$endnode color green
	set start_id [$startnode id]
	set end_id [$endnode id]
	set tfrc_agent($connection_number) [new Agent/TFRC]
	$tfrc_agent($connection_number) set fid_ $connection_number
	$ns attach-agent $startnode $tfrc_agent($connection_number)
	set vbr($vbr_in_use) [new Application/Traffic/eraVbrTrace]
	$vbr($vbr_in_use) attach-agent $tfrc_agent($connection_number)
	
	$tfrc_agent($connection_number) set packetSize_ $opt(tfrc_pktsize) ;# this is the MTU for the TFRC
	$tfrc_agent($connection_number) set TOS_field_ 1     		;# New 120905: tag ECF enabled sources! 0 is default.

	if {$connection_number < 15} {
		# Connect a file name to TFRC source to write transmit trace data:
		$tfrc_agent($connection_number) set_filename $opt(out_dir)/sd_be.tfrc.$connection_number 
	}

	$vbr($vbr_in_use) set running_ 0
	$vbr($vbr_in_use) set r_ $opt(vbr_rate) ;# Set the rate instead of packet interval
	puts2 "r_ = $opt(vbr_rate)"
	$vbr($vbr_in_use) attach-tracefile $trace_file
	$vbr($vbr_in_use) set b_ 1.5
	$vbr($vbr_in_use) set q_ 4
	$vbr($vbr_in_use) set GoP_ $opt(GoP_size)
	$vbr($vbr_in_use) set fps_ $opt(fps)
	$vbr($vbr_in_use) set isTFRC_ 1

	set tfrcsink($connection_number) [new Agent/TFRCSink]
	$ns attach-agent $endnode  $tfrcsink($connection_number)
	$ns connect $tfrc_agent($connection_number) $tfrcsink($connection_number)

	if {$connection_number == 0} {
		# Connect a file name to TFRC sink to write receivce trace data:
		$tfrcsink($connection_number) set_trace_filename $opt(out_dir)/rd_be.tfrc.$connection_number
	}
	set bwt [open $opt(out_dir)/bwt.TFRC.$connection_number.tr w] ;# $opt(bwtDelimiter).
	$ns at 0.0 "record $tfrcsink($connection_number) $bwt"
	puts2 "added TFRC connection number $connection_number (vbr($vbr_in_use)) between nodes $start_id and $end_id"
	incr vbr_in_use
}

# ======================================================================
# The interface between TFRC Receive (of ACKS)
# and the adaptive rate controller of the VBR application
# ======================================================================
Agent/TFRC instproc tfrc_ra {bytes_per_sec backlog} {
	global vbr ns opt sn_in_use
	set now [$ns now]
	$self instvar node_
	set node_id [$node_ id]
	# the following is needed to find correct value for $vbr_number,
	# which is the index for the server instance (not the physical node, 
	# but the agent) that is to receive the TFRC ack-packets:
	set vbr_number [expr $node_id + $opt(nofUDP) - 2] ;# adding values for allocated udp-sources, subtracting the core nodes (numbered 0 and 1)
	if { $opt(nofFTP) > 0 } {
		set vbr_number [expr $vbr_number - 1]
	}
	if { $opt(nofWEB) > 0 } {
		set vbr_number [expr $vbr_number - 1]
	}
	if { $opt(nofUDP) > 0 } {
		set vbr_number [expr $vbr_number - 1]
	}
	puts2 "tfrc_ra behandler ack til vbr($vbr_number) (node($node_id)) B/s: $bytes_per_sec"
	$ns at [expr $now] "$vbr($vbr_number) TFRC_rateadapt $bytes_per_sec $node_id $backlog"
}

# ======================================================================
# Function neede to receive pings properly.
# ======================================================================
Agent/Ping instproc recv {from seq ff bf} {
	$self instvar node_
	puts "node [$node_ id] received ping answer from \
		$from seq $seq with forward delay $ff and backward $bf ms."
#	puts $rttf "$seq [expr $ff/1000] [expr $bf/1000]"  ;# for RTT-logging, this can be used
}

# ==================== END PROCEDURES ==================================
# ==================== START EXEC: ENVIRONMENT =========================

set ns [new Simulator]

getopt $argc $argv

#set opt(bwtDelimiter)	-$opt(pathlossExp)-$opt(std_db)-$opt(nodeDist)m-$opt(ecn)ecn	;# variable(s) to index the bwt-files

puts2 "\n--------new sim---------"

if {$opt(seed) > -1} {
	puts2 "Seeding Random number generator with $opt(seed)"
	ns-random $opt(seed)
}
#
# Create trace
#
$ns use-newtrace
if { $opt(doTrace) } {
	set tf [open $opt(tf) w]
	$ns trace-all $tf
}
if {$opt(nam)} {
	#Open the files for nam trace
	set nf [open $opt(namtf) w]
	$ns namtrace-all $nf
	#$ns namtrace-all-wireless $nf $opt(x) $opt(y)
}

#defining colors for better readability
$ns color 40 red
$ns color 41 blue
$ns color 42 green
$ns color 43 chocolate

#
# RED and TCP parameters
#
if {$opt(ecn) == 1} {
	set opt(qm) RED
	set opt(ifq) Queue/RED
	Queue/RED set setbit_ true
}
if {$opt(qm) == "RED" } {
	Queue/RED set summarystats_ true
	Queue/RED set adaptive_ $opt(adaptive)
	Queue/RED set q_weight_ 0.0
	Queue/RED set thresh_ $opt(minth)
	Queue/RED set maxthresh_ $opt(maxth)
}
if {$opt(qm) == "DropTail" } {
	#Queue/DropTail set summarystats_ true
	#Queue/DropTail set shrink_drops_ true
}

if {$opt(nofTCP) > 0 } {
	Agent/TCP set ecn_ $opt(ecn) 
	Agent/TCP set packetSize_ $opt(tcp_pktsize)
	Agent/TCP set window_ $opt(window)
}

#
# Setting global TFRC defaults:
#
if {$opt(nofTFRC) > 0 } {
	Agent/TFRC set ss_changes_ 1 		;# Added on 10/21/2004
	Agent/TFRC set slow_increase_ 1 	;# Added on 10/20/2004
	Agent/TFRC set rate_init_ 2
	Agent/TFRC set rate_init_option_ 2 	;# Added on 10/20/2004
	Agent/TFRC set SndrType_ 1 
	Agent/TFRC set oldCode_ false
	Agent/TFRC set packetSize_ $opt(tfrc_pktsize)
	Agent/TFRC set maxqueue_ 500
	Agent/TFRC set printStatus_ true
	Agent/TFRC set ecn_ $opt(ecn) 		;# Enable ECN
	#Agent/TFRC set useHeaders_ false 	;# "useHeaders_" is removed from the evalvid adjusted tfrc.cc (TH)
	#Agent/TFRCSink set NumFeedback_ $opt(tfrcFB)
}

#
# Create wlan topology
#
setnofnodes		;# adjusts total nof servers and mobile nodes according to specified nof independent streams
wlan_topo_complex	;# sets up the topology with chosen amount of streams, one server for each source type
			;# and one mobile node for each stream

set mn_in_use 0
set sn_in_use 0
set vbr_in_use 0

if { $opt(nofTFRC)>0 || $opt(nofUDP)>0 } {
	prepare_evalvid_trace
}

#
# Set up forward FTP connection between server and mobile node
#
if {$opt(nofFTP) > 0} {
	for {set i 0} {$i < $opt(nofFTP)} {incr i} {
		attach_FTP_traffic $sn($sn_in_use) $mn($mn_in_use) $i
		incr mn_in_use
	}
	incr sn_in_use
}
#
# Set up forward WEB connection between server and mobile node
#
if {$opt(nofWEB) > 0} {
	PagePool/WebTraf set FID_ASSIGNING_MODE_ 2
	# gen/ns_tcl.cc, for 2, use sameFid_ 
	# for 1, increment FID.
	# for 0, use id_.
	set pool [new PagePool/WebTraf]
	$pool set-num-client $opt(nofWEB)
	$pool set-num-server 1	;# all clients connect to the same web-server for simplicity
	# $pool set sameFid_ $count
	$pool set sameFid_ 2
	for {set i 0} {$i < $opt(nofWEB)} {incr i} {
		attach_WEB_traffic $sn($sn_in_use) $mn($mn_in_use) $i
		incr mn_in_use
	}
	incr sn_in_use
}

#
# Set up forward UDP connection between server and mobile node
#
if {$opt(nofUDP) > 0} {
	for {set i 0} {$i < $opt(nofUDP)} {incr i} {
		attach_UDP_traffic $sn($sn_in_use) $mn($mn_in_use) $i
		incr mn_in_use
	}
	incr sn_in_use
}

#
# Set up forward TFRC connection
#
if {$opt(nofTFRC) > 0} {
	for {set i 0} {$i < $opt(nofTFRC)} {incr i} {
		attach_TFRC_traffic $sn($sn_in_use) $mn($mn_in_use) $i ;# $bs(0)
		incr mn_in_use
		incr sn_in_use		;# each TFRC session needs a seperat source to allow seperat video-feeds
	}
}


# ======================================================================
# Starting and stopping sources
# ======================================================================
if {$opt(doPing)==1} {
	ping
}
for {set i 0} { $i < $opt(nofFTP) } {incr i} {
	# start ftp sources
	$ns at [expr 10] "$ftp_server($i) start"
	$ns at [expr 50] "$ftp_server($i) stop"
}
#
# start vbr sources
#

# start UDPs
#$ns at 9.98 "$vbr(0) start"
for {set i 0} { $i < $opt(nofUDP) } {incr i} {
	$ns at 10 "$vbr($i) start"
	#$ns at 50 "$vbr($i) stop"
}
# start first TFRC
#$ns at 9.8 "$vbr($opt(nofUDP)) start"
# start the rest of the TFRCs
for {set i $opt(nofUDP)} { $i < $vbr_in_use } {incr i} {
	$ns at 10 "$vbr($i) start"
	#$ns at 50 "$vbr($i) stop"
}

$ns at $opt(stop) "stop"
puts2 ""
puts2 "-----Starting simulation!-----"
$ns run
