#!/bin/bash
#############################################################################
# Script to execute ffmpeg and generate many versions of VBR open loop video.
# Input arguments: 
# $1: Quantizer minimum.
# $2: Quantizer maximum
# $3: input file name (the video file that ffmpeg knows how to decode, i.e. trascode)
#
# Output is a *.m4v file (mpeg4 video bit stream) file
#
# Then this file is stripped for eventially repeated headers using 
#            mp4 -strip
# Then the send trace is generated with 
#            mp4 -send
# from Evalvid.
#
# Last edited by Torgeir Haukaas - 2006
############################################################################

######################################
# ffmpeg parameters:
# (Note that parameters can reflect input and output files. The file the switch belongs
# to is the first file mentioned right of the switch.)
# -r is fps, and is set twice: forces 30fps of input, and 30fps of output
# -s is set to CIF
# -vcodec is set to mpeg4 (format of output)
# -qscale is fixed quantizer (==> VBR open loop)
# -g is GOP size in numb. of frames. Note that ffmpeg will modify this to 
#    shorter GOPs when needed.
# -y forces output file overwrite in case it exists.
#####
# mp4 parameters:
# -strip searches input file for repeated headers and removes them for output
# -send <stream dest. IP address> <port numb> <fps> <source file> > <stream data>
# Note: I have speeded up this function by indicating very high frame rate!! The time
# schedule will nevertheless be overruled by the ns-2 program.
######################################
for ((  i=$1 ;  i<=$2;  i++  ))
do
	ffmpeg -r 30 -s 176x144 -i $3 -vcodec mpeg4 -g 12 -sc_threshold 20000 -qscale $i -s 176x144 -r 30 -y videoVariants/$3_Q$i.m4v
	mp4 -send 224.168.1.100 1234 1000 30000 videoVariants/$3_Q$i.m4v > nsin/st_$3_Q$i.txt
done

exit 0
